# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for logging errors, warnings, and debug statements."""

import logging
import os
import sys
import time

from .config import get_config

logging.Formatter.converter = time.gmtime

app_log = logging.getLogger(__name__)

log_level = get_config("sdk.log_level").upper()
app_log.setLevel(log_level)

# Set the format
stream_formatter = logging.Formatter(
    "[%(asctime)s] " + "[%(levelname)s] " + "%(filename)s: " + "Line "
    "%(lineno)s in %(funcName)s: " + "%(message)s"
)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(log_level)
stream_handler.setFormatter(stream_formatter)
stream_handler.stream = sys.stdout

app_log.addHandler(stream_handler)
app_log.propagate = False

# Show stack traces
log_stack_info = os.environ.get("LOGSTACK", "TRUE").upper() == "TRUE"
