# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB Config"""

from sqlalchemy.ext.declarative import declarative_base

from covalent_dispatcher._db.datastore import DataStore

Base = declarative_base()
engine = DataStore.factory().engine


def init_db(db_path: str = None):
    global engine
    engine = (
        DataStore(db_URL=db_path, initialize_db=True).engine
        if db_path is not None
        else DataStore().engine
    )
