(this.webpackJsonpwebapp=this.webpackJsonpwebapp||[]).push([[3],{710:function(t,n,e){},711:function(t,n,e){"use strict";e.r(n),e.d(n,"Terminal",(function(){return a}));var o=e(1),c=e(708),i=e(709),r=e(295),u=(e(710),e(0)),a=function(){var t=Object(o.useRef)(),n=new i.FitAddon,e=Object(r.a)(Object({NODE_ENV:"production",PUBLIC_URL:"",WDS_SOCKET_HOST:void 0,WDS_SOCKET_PATH:void 0,WDS_SOCKET_PORT:void 0,FAST_REFRESH:!0,REACT_APP_API_URL:""}).REACT_APP_SOCKET_URL,{withCredentials:!0});return Object(o.useEffect)((function(){e&&e.on("connect",(function(){var o,c;console.debug("socket ".concat(e.id," terminal connected: ").concat(e.connected)),n.fit();var i,r={cols:null===t||void 0===t||null===(o=t.current)||void 0===o?void 0:o.terminal.cols,rows:null===t||void 0===t||null===(c=t.current)||void 0===c?void 0:c.terminal.rows};r.rows&&(null===t||void 0===t||null===(i=t.current)||void 0===i||i.terminal.focus(),e.emit("start_terminal",r),e.emit("resize",r))}))})),Object(o.useEffect)((function(){if(e)return function(){e.disconnect()}})),Object(o.useEffect)((function(){e&&e.on("pty-output",(function(n){var e;null===t||void 0===t||null===(e=t.current)||void 0===e||e.terminal.write(n.output)}))})),Object(u.jsx)(u.Fragment,{children:e&&Object(u.jsx)(c.a,{"data-testid":"terminal",options:{cursorBlink:!0,macOptionIsMeta:!0,scrollback:500,theme:{background:"#08081A",foreground:"#AEB6FF"}},className:"term",ref:t,onData:function(t){e.emit("pty-input",{input:t})},addons:[n]})})};n.default=Object(o.memo)(a)}}]);
//# sourceMappingURL=3.6e4b72b9.chunk.js.map